type bool = True | False

type elt = Abstract_element

type cmp = Lt | Gt | Eq

val compare: 'a -> 'a -> cmp

type redblack =
  | Black of redblack * elt * redblack
  | Red of redblack * elt * redblack
  | Empty

let rec mem x = function
  | Empty -> False
  | (Red(l, y, r)
    | Black(l, y, r)) ->
    match compare x y with
    | Lt -> mem x l
    | Gt -> mem x r
    | Eq -> True

let singleton x =
  Black(Empty, x, Empty)

let balance_left l x r =
  match (l, x, r) with
  | ((Red(Red(a, x, b), y, c), z, d)
    | (Red(a, x, Red(b, y, c)), z, d)) ->
    Red(Black(a, x, b), y, Black(c, z, d))
  | (l, x, r) ->
    Black(l, x, r)

let balance_right l x r =
  match (l, x, r) with
  | ((a, x, Red(Red(b, y, c), z, d))
    | (a, x, Red(b, y, Red(c, z, d)))) ->
    Red(Black(a, x, b), y, Black(c, z, d))
  | (l, x, r) ->
    Black(l, x, r)

let add x s =
  let rec add_aux t = match t with
    | Empty ->
      Red(Empty, x, Empty)
    | Red(l, y, r) ->
      begin match compare x y with
        | Lt -> Red(add_aux l, y, r)
        | Gt -> Red(l, y, add_aux r)
        | Eq -> t end
    | Black(l, y, r)  ->
      begin match compare x y with
        | Lt -> balance_left (add_aux l) y r
        | Gt -> balance_right l y (add_aux r)
        | Eq -> t
      end
  in
  match add_aux s with
  | Red(a, b, c) -> Black(a, b, c)
  | a -> a



type tree =
  | SEmpty
  | SNode of tree * elt * tree


type relation simplification: redblack => tree with
  | Empty => SEmpty
  | (Red(a, e, b) | Black(a, e, b)) => SNode(a, e, b) when a b: simplification


let simpl_mem = lifting mem: elt -> simplification -> bool

let simpl_singleton = lifting singleton: elt -> simplification

let simpl_add = lifting add: elt -> simplification -> simplification


let simpl_add = lifting add: elt -> simplification -> simplification with
| (match #a with Left _ -> _a) => #a <- Left ()
