type binop' =
  | Add'
  | Mul'

type expr' =
  | Const' of int
  | Binop' of binop' * expr' * expr'
  | Abs' of (expr' -> expr' option)
  | App' of expr' * expr'

let rec eval' e = match e with
    | Const' i -> Some (Const' i)
    | Binop'(Add', u, v) ->
      begin match (eval' u, eval' v) with
        | (Some (Const' i), Some (Const' i1)) ->
          Some (Const' (add i i1))
        | _ -> None
      end
    | Binop'(Mul', u, v) ->
      begin match (eval' u, eval' v) with
        | (Some (Const' i), Some (Const' i1)) ->
          Some (Const' (mul i i1))
        | _ -> None
      end
    | Abs' f -> Some (Abs' f)
    | App'(u, v) ->
      begin match eval' u with
        | Some (Abs' f) ->
          begin match eval' v with
            | None -> None
            | Some x -> f x
          end
        | _ -> None
      end

type expr'' =
  | Const'' of int
  | Binop'' of binop' * expr'' * expr''
  | Abs'' of (value'' -> value'' option)
  | App'' of expr'' * expr''
and value'' =
  | Int'' of int
  | Fun'' of (value'' -> value'' option)

let rec eval'' e = match e with
    | Const'' i -> Some (Int'' i)
    | Binop''(Add', u, v) ->
      begin match (eval'' u, eval'' v) with
        | (Some (Int'' i), Some (Int'' i1)) ->
          Some (Int'' (add i i1))
        | _ -> None
      end
    | Binop''(Mul', u, v) ->
      begin match (eval'' u, eval'' v) with
        | (Some (Int'' i), Some (Int'' i1)) ->
          Some (Int'' (mul i i1))
        | _ -> None
      end
    | Abs'' f -> Some (Fun'' f)
    | App''(u, v) ->
      begin match eval'' u with
        | Some (Fun'' f) ->
          begin match eval'' v with
            | None -> None
            | Some x -> f x
          end
        | _ -> None
      end
