let rec eval' env e = match e with
    | Var x -> assoc x env
    | Abs(x, e) -> Some (VAbs(x, e))
    | App(e1, e2) ->
      begin match eval' env e1 with
        | (None | Some (VTup _)) -> None
        | Some (VAbs(x, e)) ->
          begin match eval' env e2 with
            | None -> None
            | Some v -> eval' (Cons((x, v), env)) e
          end
      end
    | Tup e ->
      begin match eval' env e with
        | None -> None
        | Some e -> Some (VTup e)
      end
    | Proj e ->
      begin match eval' env e with
        | (None | Some (VAbs(_, _))) -> None
        | Some (VTup e) -> Some e
      end

type value' =
  | VClosure' of id * (id * value') list * expr
  | VTup' of value'

let rec eval'' env e = match e with
    | Var x -> assoc x env
    | Abs(x, e) -> Some (VClosure'(x, env, e))
    | App(e1, e2) ->
      begin match eval'' env e1 with
        | (None | Some (VTup' _)) -> None
        | Some (VClosure'(x, _, e)) ->
          begin match eval'' env e2 with
            | None -> None
            | Some v -> eval'' (Cons((x, v), env)) e
          end
      end
    | Tup e ->
      begin match eval'' env e with
        | None -> None
        | Some v -> Some (VTup' v)
      end
    | Proj e ->
      begin match eval'' env e with
        | (None | Some (VClosure'(_, _, _))) -> None
        | Some (VTup' v) -> Some v
      end

let rec eval'' env e = match e with
    | Var x -> assoc x env
    | Abs(x, e) -> Some (VClosure'(x, #48, e))
    | App(e1, e2) ->
      begin match eval'' env e1 with
        | (None | Some (VTup' _)) -> None
        | Some (VClosure'(x, _, e)) ->
          begin match eval'' env e2 with
            | None -> None
            | Some v -> eval'' (Cons((x, v), env)) e
          end
      end
    | Tup e ->
      begin match eval'' env e with
        | None -> None
        | Some v -> Some (VTup' v)
      end
    | Proj e ->
      begin match eval'' env e with
        | (None | Some (VClosure'(_, _, _))) -> None
        | Some (VTup' v) -> Some v
      end

let rec eval'' env e = match e with
    | Var x -> assoc x env
    | Abs(x, e) -> Some (VClosure'(x, env, e))
    | App(e1, e2) ->
      begin match eval'' env e1 with
        | (None | Some (VTup' _)) -> None
        | Some (VClosure'(x, _, e)) ->
          begin match eval'' env e2 with
            | None -> None
            | Some v -> eval'' (Cons((x, v), env)) e
          end
      end
    | Tup e ->
      begin match eval'' env e with
        | None -> None
        | Some v -> Some (VTup' v)
      end
    | Proj e ->
      begin match eval'' env e with
        | (None | Some (VClosure'(_, _, _))) -> None
        | Some (VTup' v) -> Some v
      end

type expr' =
  | Var' of id
  | Abs' of id * expr'
  | Closure' of id * (id * expr') list * expr'
  | App' of expr' * expr'
  | Tup' of expr'
  | Proj' of expr'

let rec eval' env e = match e with
    | Var' x -> assoc x env
    | Abs'(x, e) -> Some (Closure'(x, env, e))
    | Closure'(x, _, e) -> Some (Closure'(x, env, e))
    | App'(e1, e2) ->
      begin match eval' env e1 with
        | (None | Some (Tup' _)) -> None
        | Some (Abs'(x, e)) ->
          begin match eval' env e2 with
            | None -> None
            | Some v -> eval' (Cons((x, v), env)) e
          end
        | Some (Closure'(x, _, e)) ->
          begin match eval' env e2 with
            | None -> None
            | Some v -> eval' (Cons((x, v), env)) e
          end
        | ((Some (Var' _) | Some (App'(_, _)))
          | Some (Proj' _)) -> fail ()
      end
    | Tup' e ->
      begin match eval' env e with
        | None -> None
        | Some e -> Some (Tup' e)
      end
    | Proj' e ->
      begin match eval' env e with
        | ((None | Some (Abs'(_, _)))
          | Some (Closure'(_, _, _))) -> None
        | Some (Tup' e) -> Some e
        | ((Some (Var' _) | Some (App'(_, _)))
          | Some (Proj' _)) -> fail ()
      end

let rec eval'' env e = match e with
    | Var x -> assoc x env
    | Abs(x, e) -> Some (VClosure'(x, env, e))
    | App(e1, e2) ->
      begin match eval'' env e1 with
        | (None | Some (VTup' _)) -> None
        | Some (VClosure'(x, env1, e)) ->
          begin match eval'' env e2 with
            | None -> None
            | Some v -> eval'' (Cons((x, v), env1)) e
          end
      end
    | Tup e ->
      begin match eval'' env e with
        | None -> None
        | Some e -> Some (VTup' e)
      end
    | Proj e ->
      begin match eval'' env e with
        | (None | Some (VClosure'(_, _, _))) -> None
        | Some (VTup' e) -> Some e
      end
