type expr =
  | Var of string
  | Abs of string * expr
  | App of expr * expr
  | Const of int

type ('a,'err) result =
  | Ok of 'a
  | Error of 'err

let rec eval_mes_ env e = match e with
    | Var x ->
      begin match find x env with
        | Some a -> Ok a
        | None -> Error #30
      end
    | App(u, v) ->
      begin match eval_mes_ env u with
        | Ok (Abs(x, f)) ->
          begin match eval_mes_ env v with
            | Ok a -> eval_mes_ (Cons((x, a), env)) f
            | Error _ -> Error #14
          end
        | Ok (Var _) -> Error #22
        | Ok (App(_, _)) -> Error #12
        | Ok (Const _) -> Error #10
        | Error _ -> Error #8
      end
    | (Abs(_, _) | Const _) -> Ok e

let rec eval_mes env e = match e with
    | Var x ->
      begin match find x env with
        | Some a -> Ok a
        | None -> Error str__Unbound_variable
      end
    | App(u, v) ->
      begin match eval_mes env u with
        | Ok (Abs(x, f)) ->
          begin match eval_mes env v with
            | Ok a -> eval_mes (Cons((x, a), env)) f
            | Error x -> Error x
          end
        | ((Ok (Var _) | Ok (App(_, _))) | Ok (Const _)) ->
          Error str__Not_a_function
        | Error x -> Error x
      end
    | (Abs(_, _) | Const _) -> Ok e

type location = Location of string * int * int

type expr' =
  | Var' of string
  | Abs' of string * (expr' * location)
  | App' of (expr' * location) * (expr' * location)
  | Const' of int

type ('a,'err) result' =
  | Ok' of 'a
  | Error' of 'err

let rec _eval_loc env e = match e with
    | (Var' x, _) ->
      begin match find x env with
        | Some a -> Ok a
        | None -> Error (str__Unbound_variable, #33)
      end
    | (App'(u, v), _) ->
      begin match _eval_loc env u with
        | Ok (Abs'(x, f), _) ->
          begin match _eval_loc env v with
            | Ok a -> _eval_loc (Cons((x, a), env)) f
            | Error (b, _) -> Error (b, #16)
          end
        | Ok (Var' _, _) -> Error (str__Not_a_function, #24)
        | Ok (App'(_, _), _) ->
          Error (str__Not_a_function, #13)
        | Ok (Const' _, _) ->
          Error (str__Not_a_function, #10)
        | Error (b, _) -> Error (b, #8)
      end
    | ((Abs'(_, _), _) | (Const' _, _)) -> Ok e

let rec eval_loc env e = match e with
    | (Var' x, loc) ->
      begin match find x env with
        | Some a -> Ok a
        | None -> Error (str__Unbound_variable, loc)
      end
    | (App'(u, v), _) ->
      begin match eval_loc env u with
        | Ok (Abs'(x, f), _) ->
          begin match eval_loc env v with
            | Ok a -> eval_loc (Cons((x, a), env)) f
            | Error (b, x) -> Error (b, x)
          end
        | ((Ok (Var' _, x) | Ok (App'(_, _), x))
          | Ok (Const' _, x)) ->
          Error (str__Not_a_function, x)
        | Error (b, x) -> Error (b, x)
      end
    | ((Abs'(_, _), _) | (Const' _, _)) -> Ok e

let rec _eval_loc env e = match e with
    | (Var' x, _) ->
      begin match find x env with
        | Some a -> Ok a
        | None -> Error str__Unbound_variable
      end
    | (App'(u, v), _) ->
      begin match _eval_loc env u with
        | Ok (Abs'(x, f), _) ->
          begin match _eval_loc env v with
            | Ok a -> _eval_loc (Cons((x, a), env)) f
            | Error x -> Error x
          end
        | ((Ok (Var' _, _) | Ok (App'(_, _), _))
          | Ok (Const' _, _)) -> Error str__Not_a_function
        | Error x -> Error x
      end
    | ((Abs'(_, _), _) | (Const' _, _)) -> Ok e

let rec free_vars_loc e = match e with
    | (Var' x, _) -> Cons(x, Nil)
    | (Abs'(x, f), _) -> remove x (free_vars_loc f)
    | (App'(u, v), _) ->
      append (free_vars_loc u) (free_vars_loc v)
    | (Const' _, _) -> Nil
