
type expr =
  | Var of string
  | Abs of string * expr
  | App of expr * expr
  | Const of int

let rec eval env e = match e with
  | Var x ->
      begin match find x env with
      | Some v -> Some v
      | None -> None
      end
  | App (u, v) ->
      begin match eval env u with
      | Some (Abs (x, f)) ->
          begin match eval env v with
          | Some w -> eval (Cons ((x, w), env)) f
          | None -> None
          end
      | Some w -> None
      | None -> None
      end
  | v -> Some v

type ('a,'err) result =
  | Ok of 'a
  | Error of 'err

type ornament ('a, 'err) optres : 'a option => ('a, 'err) result with
  | Some a => Ok a
  | None => Error _

let eval_mes_ =
  lifting eval : _ -> _ -> ('a, 'err) optres with
  ornament * <- @id

let eval_mes =
  lifting eval : _ -> _ -> ('a, 'err) optres with
  ornament * <- @id
  patch #30[str__Unbound_variable]
  patch (match _a with Ok _ -> Error (#[str__Not_a_function]))
  patch (match _a with Error a -> Error (#[a]))

type location = Location of string * int * int

type expr' =
  | Var' of string
  | Abs' of string * (expr' * location)
  | App' of (expr' * location) * (expr' * location)
  | Const' of int

type ornament add_loc : expr => expr' * location with
  | Var x => (Var' x, _)
  | Abs (x, f) => (Abs' (x, f), _) when f : add_loc
  | App (u, v) => (App' (u, v), _) when u v : add_loc
  | Const i => (Const' i, _)


type ('a,'err) result' =
  | Ok' of 'a
  | Error' of 'err

type ornament ('a, 'mes, 'err) optres' : ('a, 'mes) result
        => ('a, 'mes * 'err) result with
  | Ok a => Ok a
  | Error b => Error (b, _)

let _eval_loc =
  lifting eval_mes :
     (_ * add_loc) list -> add_loc -> (_, _, _) optres'
  with ornament * <- @id

let eval_loc =
  lifting eval_mes : (_ * add_loc) list -> add_loc
        -> (_, _, _) optres' with
  ornament * <- @id
  patch #33[begin match e with (_, loc) -> loc (* unbound *) end]
  patch match _a with Ok(_, _b) -> Error (_c, (#[_b]))
  patch match _a with Error(_, _b) -> Error (_c, (#[_b]))

 let _eval_loc =
  lifting eval_mes : (_ * add_loc) list -> add_loc -> _ with
   ornament * <- @id

let rec free_vars e = match e with
  | Var x -> Cons (x, Nil)
  | Abs (x, f) -> remove x (free_vars f)
  | App (u, v) -> append (free_vars u) (free_vars v)
  | Const i -> Nil

let free_vars_loc = lifting free_vars : add_loc -> string list
    with ornament * <- @id
