
type expr =
  | Abs of (expr -> expr)
  | App of expr * expr
  | Const of int

let rec eval env e =
  match e with
  | App (u, v) ->
      let res_u = eval env u in
      begin match res_u with
      | Some (Abs f) -> eval env (f v)
      | _ -> None
      end
  | _ -> Some(e)

type expr_big =
  | Absb of (expr_big -> expr_big)
  | Appb of expr_big * expr_big
  | Letb of string * expr_big * expr_big
  | Constb of int
  | Varb of string
  | Addb of expr_big * expr_big

let rec eval_big env e = match e with
  | (Absb' _, _) -> Ok e
  | (Appb'(u, v), loc) ->
      let res_u = eval_big env u in
      begin match res_u with
      | Ok (Absb' f, _) -> eval_big env (f v)
      | Ok _ -> Error loc
      | Error x -> Error x
      end
  | (Addb' (u, v), loc) ->
      let u' = eval_big env u in
      let v' = eval_big env v in
      begin match u' with
      | Ok (Constb' i, _) ->
          begin match v' with
          | Ok (Constb' i1, _) ->
            Ok (Constb' (plus i i1), loc)
          | Error x -> Error x
          | Ok (_, _) -> Error loc
          end
      | Ok (_, _) -> Error loc
      | Error x -> Error x
      end
  | (Constb' _, _) -> Ok e
  | (Varb' i, loc) ->
      begin match assoc i env with
      | Some a -> Ok a
      | None -> Error loc
      end
  | (Letb'(s, u, v), _) ->
      begin match eval_big env u with
      | Ok a -> eval_big (Cons((s, a), env)) v
      | Error x -> Error x
      end
