type expr' =
  | Abs' of ((expr' * location) -> (expr' * location))
  | App' of (expr' * location) * (expr' * location)
  | Const' of int

let rec _eval_incomplete env e = match e with
    | (App'(u, v), _) ->
      let res_u = _eval_incomplete env u in
      begin match res_u with
        | Ok (Abs' f, _) -> _eval_incomplete env (f v)
        | Ok (App'(_, _), _) -> Error #8
        | Ok (Const' _, _) -> Error #6
        | Error _ -> Error #13
      end
    | _ -> Ok e

let rec eval' env e = match e with
    | (App'(u, v), loc) ->
      let res_u = eval' env u in
      begin match res_u with
        | Ok (Abs' f, _) -> eval' env (f v)
        | (Ok (App'(_, _), _) | Ok (Const' _, _)) ->
          Error loc
        | Error x -> Error x
      end
    | _ -> Ok e

let rec eval' env e = match e with
    | (App'(u, v), loc) ->
      let res_u = eval' env u in
      begin match res_u with
        | Ok (Abs' f, _) -> eval' env (f v)
        | (Ok (App'(_, _), _) | Ok (Const' _, _)) ->
          Error loc
        | Error x -> Error x
      end
    | _ -> Ok e

type expr2 =
  | Abs2 of (expr2 -> expr2)
  | App2 of expr2 * expr2
  | Const2 of int
  | Add2 of expr2 * expr2

let rec eval2 env e = match e with
    | App2(u, v) ->
      let res_u = eval2 env u in
      begin match res_u with
        | Some (Abs2 f) -> eval2 env (f v)
        | ((None | Some (App2(_, _))) | Some (Const2 _)) ->
          None
        | Some (Add2(_, _)) -> #1
      end
    | (Abs2 _ | Const2 _) -> Some e
    | Add2(_, _) -> #2

let rec eval2 env e = match e with
    | App2(u, v) ->
      let res_u = eval2 env u in
      begin match res_u with
        | Some (Abs2 f) -> eval2 env (f v)
        | _ -> None
      end
    | (Abs2 _ | Const2 _) -> Some e
    | Add2(x, x1) ->
      let u' = eval2 env x in
      let v' = eval2 env x1 in
      begin match u' with
        | Some (Const2 i) ->
          begin match v' with
            | Some (Const2 j) -> Some (Const2 (plus i j))
            | (((None | Some (Abs2 _)) | Some (App2(_, _)))
              | Some (Add2(_, _))) -> None
          end
        | (((None | Some (Abs2 _)) | Some (App2(_, _)))
          | Some (Add2(_, _))) -> None
      end

type expr2' =
  | Abs2' of ((expr2' * location) -> (expr2' * location))
  | App2' of (expr2' * location) * (expr2' * location)
  | Add2' of (expr2' * location) * (expr2' * location)
  | Const2' of int

let rec eval2' env e = match e with
    | (App2'(u, v), _) ->
      let res_u = eval2' env u in
      begin match res_u with
        | Ok (Abs2' f, _) -> eval2' env (f v)
        | Ok (App2'(_, _), _) -> Error #43
        | Ok (Add2'(_, _), _) -> Error #39
        | Ok (Const2' _, _) -> Error #41
        | Error _ -> Error #48
      end
    | ((Abs2' _, _) | (Const2' _, _)) -> Ok e
    | (Add2'(u, v), _) ->
      let u' = eval2' env u in
      let v' = eval2' env v in
      begin match u' with
        | Ok (Const2' i, _) ->
          begin match v' with
            | Ok (Const2' i1, _) ->
              Ok (Const2' (plus i i1), #12)
            | Ok (Abs2' _, _) -> Error #17
            | Ok (App2'(_, _), _) -> Error #15
            | Ok (Add2'(_, _), _) -> Error #8
            | Error _ -> Error #21
          end
        | Ok (Abs2' _, _) -> Error #27
        | Ok (App2'(_, _), _) -> Error #25
        | Ok (Add2'(_, _), _) -> Error #6
        | Error _ -> Error #31
      end

let rec eval2' env e = match e with
    | (App2'(u, v), loc) ->
      let res_u = eval2' env u in
      begin match res_u with
        | Ok (Abs2' f, _) -> eval2' env (f v)
        | ((Ok (App2'(_, _), _) | Ok (Add2'(_, _), _))
          | Ok (Const2' _, _)) -> Error loc
        | Error x -> Error x
      end
    | ((Abs2' _, _) | (Const2' _, _)) -> Ok e
    | (Add2'(u, v), loc) ->
      let u' = eval2' env u in
      let v' = eval2' env v in
      begin match u' with
        | Ok (Const2' i, _) ->
          begin match v' with
            | Ok (Const2' i1, _) ->
              Ok (Const2' (plus i i1), #12)
            | ((Ok (Abs2' _, _) | Ok (App2'(_, _), _))
              | Ok (Add2'(_, _), _)) -> Error loc
            | Error x -> Error x
          end
        | ((Ok (Abs2' _, _) | Ok (App2'(_, _), _))
          | Ok (Add2'(_, _), _)) -> Error loc
        | Error x -> Error x
      end

let rec eval2' env e = match e with
    | (App2'(u, v), loc) ->
      let res_u = eval2' env u in
      begin match res_u with
        | Ok (Abs2' f, _) -> eval2' env (f v)
        | ((Ok (App2'(_, _), _) | Ok (Add2'(_, _), _))
          | Ok (Const2' _, _)) -> Error loc
        | Error x -> Error x
      end
    | ((Abs2' _, _) | (Const2' _, _)) -> Ok e
    | (Add2'(u, v), loc) ->
      let u' = eval2' env u in
      let v' = eval2' env v in
      begin match u' with
        | Ok (Const2' i, _) ->
          begin match v' with
            | Ok (Const2' i1, _) ->
              Ok (Const2' (plus i i1), loc)
            | ((Ok (Abs2' _, _) | Ok (App2'(_, _), _))
              | Ok (Add2'(_, _), _)) -> Error loc
            | Error x -> Error x
          end
        | ((Ok (Abs2' _, _) | Ok (App2'(_, _), _))
          | Ok (Add2'(_, _), _)) -> Error loc
        | Error x -> Error x
      end

let rec eval_desorn env e = match e with
    | (Absb _ | Constb _) -> Some e
    | Appb(u, v) ->
      let res_u = eval_desorn env u in
      begin match res_u with
        | Some (Absb f) -> eval_desorn env (f v)
        | (((((None | Some (Appb(_, _)))
          | Some (Letb(_, _, _)))
          | Some (Constb _))
          | Some (Varb _))
          | Some (Addb(_, _))) -> None
      end
    | Addb(u, v) ->
      let u' = eval_desorn env u in
      let v' = eval_desorn env v in
      begin match u' with
        | Some (Constb i) ->
          begin match v' with
            | Some (Constb i1) -> Some (Constb (plus i i1))
            | (((((None | Some (Absb _)) | Some (Appb(_, _)))
              | Some (Letb(_, _, _)))
              | Some (Varb _))
              | Some (Addb(_, _))) -> None
          end
        | (((((None | Some (Absb _)) | Some (Appb(_, _)))
          | Some (Letb(_, _, _)))
          | Some (Varb _))
          | Some (Addb(_, _))) -> None
      end
    | Varb i ->
      begin match assoc i env with
        | Some a -> Some a
        | None -> None
      end
    | Letb(s, u, v) ->
      begin match eval_desorn env u with
        | Some a -> eval_desorn (Cons((s, a), env)) v
        | None -> None
      end

let rec eval_desorn_orn env e = match e with
    | ((Absb' _, _) | (Constb' _, _)) -> Ok e
    | (Appb'(u, v), loc) ->
      let res_u = eval_desorn_orn env u in
      begin match res_u with
        | Ok (Absb' f, _) -> eval_desorn_orn env (f v)
        | ((((Ok (Appb'(_, _), _) | Ok (Constb' _, _))
          | Ok (Varb' _, _))
          | Ok (Letb'(_, _, _), _))
          | Ok (Addb'(_, _), _)) -> Error loc
        | Error x -> Error x
      end
    | (Addb'(u, v), loc) ->
      let u' = eval_desorn_orn env u in
      let v' = eval_desorn_orn env v in
      begin match u' with
        | Ok (Constb' i, _) ->
          begin match v' with
            | Ok (Constb' i1, _) ->
              Ok (Constb' (plus i i1), loc)
            | ((((Ok (Absb' _, _) | Ok (Appb'(_, _), _))
              | Ok (Varb' _, _))
              | Ok (Letb'(_, _, _), _))
              | Ok (Addb'(_, _), _)) -> Error loc
            | Error x -> Error x
          end
        | ((((Ok (Absb' _, _) | Ok (Appb'(_, _), _))
          | Ok (Varb' _, _))
          | Ok (Letb'(_, _, _), _))
          | Ok (Addb'(_, _), _)) -> Error loc
        | Error x -> Error x
      end
    | (Varb' i, _) ->
      begin match assoc i env with
        | Some a -> Ok a
        | None -> Error #49
      end
    | (Letb'(s, u, v), _) ->
      begin match eval_desorn_orn env u with
        | Ok a -> eval_desorn_orn (Cons((s, a), env)) v
        | Error x -> Error x
      end

let rec eval_desorn_orn env e = match e with
    | ((Absb' _, _) | (Constb' _, _)) -> Ok e
    | (Appb'(u, v), loc) ->
      let res_u = eval_desorn_orn env u in
      begin match res_u with
        | Ok (Absb' f, _) -> eval_desorn_orn env (f v)
        | ((((Ok (Appb'(_, _), _) | Ok (Constb' _, _))
          | Ok (Varb' _, _))
          | Ok (Letb'(_, _, _), _))
          | Ok (Addb'(_, _), _)) -> Error loc
        | Error x -> Error x
      end
    | (Addb'(u, v), loc) ->
      let u' = eval_desorn_orn env u in
      let v' = eval_desorn_orn env v in
      begin match u' with
        | Ok (Constb' i, _) ->
          begin match v' with
            | Ok (Constb' i1, _) ->
              Ok (Constb' (plus i i1), loc)
            | ((((Ok (Absb' _, _) | Ok (Appb'(_, _), _))
              | Ok (Varb' _, _))
              | Ok (Letb'(_, _, _), _))
              | Ok (Addb'(_, _), _)) -> Error loc
            | Error x -> Error x
          end
        | ((((Ok (Absb' _, _) | Ok (Appb'(_, _), _))
          | Ok (Varb' _, _))
          | Ok (Letb'(_, _, _), _))
          | Ok (Addb'(_, _), _)) -> Error loc
        | Error x -> Error x
      end
    | (Varb' i, loc) ->
      begin match assoc i env with
        | Some a -> Ok a
        | None -> Error loc
      end
    | (Letb'(s, u, v), _) ->
      begin match eval_desorn_orn env u with
        | Ok a -> eval_desorn_orn (Cons((s, a), env)) v
        | Error x -> Error x
      end
