
module Loc = struct
type loc = Loc of string * Lexing.position * Lexing.position
let dummy_loc = Loc("<dummy>",Lexing.dummy_pos, Lexing.dummy_pos)
let lex_loc file lexbuf = Lexing.(Loc(file, lexbuf.lex_start_p, lexbuf.lex_curr_p))
let makepos s pstart pend = Loc(s,pstart,pend)
let merge_loc (Loc(s1,st,_)) (Loc(_,_,en)) = Loc(s1,st,en)
let string_of_loc (Loc(s1,p1,p2)) =
  let p l =
    Lexing.(string_of_int l.pos_lnum ^":"^ string_of_int (l.pos_cnum - l.pos_bol))
  in
  "("^s1^":"^p p1^"-"^p p2^")"
end

exception Error of Loc.loc * string

open Lexing

exception User_error of string

let print_error (Loc.Loc(f,p1,p2)) str =
  let fname = p1.pos_fname in
  let f = if f = "<dummy>" || fname = "" then f else fname in  
  Printf.eprintf "File %S, lines %d-%d, characters %d-%d:\n%s\n\n"
    f p1.pos_lnum p2.pos_lnum  (p1.pos_cnum - p1.pos_bol) (p2.pos_cnum - p1.pos_bol) str

let error loc fmt = Format.kasprintf (fun str -> raise (Error(loc,str))) fmt
let warning loc fmt =
  Format.kasprintf (print_error loc) fmt
let failwith fmt = Format.kasprintf failwith fmt
let user_error fmt = Format.kasprintf (fun x -> raise (User_error x)) fmt
