open Ast

module E = Explicit.E

exception Invalid_patch

type expr = Explicit.expr

type patch_list = (patch * loc) list

let find_best_patch (pl: patch_list) (pa: patch_addr): patch = assert false

let match_addr a patch_id = match a with
  | PatchFor(l) -> List.exists (function (_, None), _ -> true | (_, Some s), _ -> s = patch_id) l
  | DefaultPatch -> true

let gty = E.TVar(id "##UNTYPED##")

let subst_for_inst_req (oenv: Explicit.orn_env) (env: env) (liftings: int -> id * id) (orn_name_from_var: id -> id) (pl: patch_list) = function
  | Orn.InstantiateEnv(_,n,p,x_base,_,_) ->
    IdHashtbl.add Pretty.holemap p ("(lifting #"^string_of_int n^" of "^x_base.id_name^")");
    begin match liftings n with
      | _, y_lift -> [(p, (E.Var y_lift, gty, dummy_loc))], []
      | exception Not_found ->
        say "no lifting given/inferred for [lifting #%d of %a]" n Pp.id_nice x_base;
        [], []
    end
  | Orn.InstantiateOrnVar(_,n,ovar,c,_,xext,xcon,xproj) ->
    IdHashtbl.add Pretty.holemap xcon ("(orn-cons #"^string_of_int n^")");
    IdHashtbl.add Pretty.holemap xproj ("(orn-match #"^string_of_int n^")");
    begin match orn_name_from_var ovar with
      | orn_name ->
        let (orn_descr, ext, con, proj) = List.assoc orn_name oenv.Orn.env_ornaments in
        say "using ornament %a for [ornament #%d of %a]" Pp.id_nice orn_name n Pp.id_nice c;
        [(xcon, con); (xproj, proj)], [(xext,ext)]
      | exception Not_found -> say "no ornament given/inferred for [ornament #%d of %a]" n Pp.id_nice c; [], []
    end
  | Orn.IdentityOrnament(c,_,xext,xcon,xproj) ->
    let (_, ext, con, proj) = List.assoc c oenv.Orn.env_ornaments in
    [(xcon, con); (xproj, proj)], [(xext,ext)]
  | Orn.InstantiatePatch(vars,_,n,x,_) ->
    List.iter (fun x -> IdHashtbl.add Pretty.holemap x ("#"^string_of_int n)) x;
    (*begin match search pl (function PatchHole(a,e), _ when match_addr a n -> Some(e) | _ -> None) with
      | e ->
       let e = Ml_inference.fix_contuple env e in
       let e = Orn_codegen.name_patch env.scope vars e in
       let e = Desugaring.untyped_all_patch env e in
       let e = Eml_compute.ExplicitToGml.ml_to_gml e in
       [(x,e)], []
      | exception Not_found ->
       say "no patch given for [#%d] %a" n Pp.id x; [], []
      end*) [], []

exception Doesnot_fit

type impl_expr = Ast.expr

let subst_impl l e =
  Implicit.visit_expr (fun () () e ->
      (), (), match e with
      | Var(x), _ -> (try List.assoc x l with | Not_found -> e)
      | e -> e) () () e
  |> snd

let subst_expr l e =
  Explicit.visit_expr (fun () () e ->
      All (), (), match e with
      | Explicit.Var(x, _), _, _ -> (try List.assoc x l with | Not_found -> e)
      | e -> e) () () e
  |> snd

let subst_pattern l (e: Explicit.pat) =
  Explicit.visit (fun () e ->
      (), match e with
      | Explicit.PVar(x), ty, loc -> (try Explicit.PVar (List.assoc x l), ty, loc with | Not_found -> e)
      | e -> e) () e
  |> snd

let rec pat_to_impl (p: Explicit.pat) =
  let p', _, l = p in
  begin match p' with
    | Ast.Explicit.PCon (i, _, p) -> Con(FromId i, List.map pat_to_impl p)
    | Ast.Explicit.PVar i -> Var(FromId i)
    | Ast.Explicit.(PWild | PAlt(_, _) | PZero) -> Format.eprintf "wildcards/alt/zero not supported@."; raise Doesnot_fit
  end, l

let rec pat_to_impl_pat (p: Explicit.pat): pat =
  let p', _, l = p in
  begin match p' with
    | Explicit.PCon(i, _, p) -> PCon(FromId i, List.map pat_to_impl_pat p), l
    | Explicit.PVar i -> PVar (FromId i), l
    | Explicit.PWild -> PWild, l
    | Explicit.PAlt(a, b) -> PAlt (pat_to_impl_pat a, pat_to_impl_pat b), l
    | Explicit.PZero -> PZero, l
  end

let rec to_impl_expr (e:expr): impl_expr =
  let e', _, l = e in
  begin match e' with
    | Absurd -> Patch (PatchFor []), l
    | Ast.Explicit.Var (a, _) -> Var(FromId a), l
    | Ast.Explicit.Let (a, b, _, iddef, e) ->
      Let(a, b, List.map (fun (a, _, b) ->
          FromId a, to_impl_expr b) iddef, to_impl_expr e), l
    | Ast.Explicit.Match (a, cl) ->
      Match(to_impl_expr a, List.map (fun (a, b) -> pat_to_impl_pat a, to_impl_expr b) cl), l
    | Ast.Explicit.Con (a, _, b) -> Con(FromId a, List.map (fun e -> to_impl_expr e) b), l
    | Ast.Explicit.Abs (a, _, b) -> Abs (FromId a, to_impl_expr b), l
    | Ast.Explicit.App (a, b) -> App (to_impl_expr a, to_impl_expr b), l
    | Ast.Explicit.Mark (m, e) -> to_impl_expr e
    | Ast.Explicit.MultiFix (funs, fun_name) -> Let(Ast.IsMeta, true, List.map (fun (fun_id, (var_name, var_ty), e) ->
        FromId fun_id, (Abs(FromId var_name, to_impl_expr e), l)) funs, (Var(FromId fun_name), l)), l
    | Ast.Explicit.ShallowMatch _ -> assert false
    | Ast.Explicit.ShallowCon (_, _, _, _) -> assert false
    | Patch a -> Patch a, l
  end

let rec chop n l = match n, l with
  | 0, _ -> [], l
  | n, t::q ->
    let a, b = chop (n-1) q in
    t::a, b
  | _ -> assert false

let mem_var (var:id) (e:expr) =
  Ast.Explicit.visit_expr (fun () c e ->
      All (), (match e with
          | Explicit.Var (i, _), _, _ when i = var -> true
          | _ -> c), e) () false e |> fst

let rec try_unify (env: env) (target:id) (e_:expr) (e'_: impl_expr) =
  let e, ty_e, _ = e_ in
  let e', _ = e'_ in
  let tu = try_unify env target in

  match e, e' with
  | Explicit.Var(i, _), Patch(_) -> []
  | ShallowCon(_), _ | ShallowMatch _, _ -> assert false
  | Mark(_, e), e' -> tu e e'_
  | Explicit.Match(x, cl), Implicit.Match(x', [a]) ->
    if mem_var target x then
      let l = tu x x' in
      l @ try
        search cl (fun cl ->
            try
              let l = unify_pat (fst cl) (fst a) in
              Some (l @ tu (snd cl) (subst_impl l (snd a)))
            with
            | Doesnot_fit -> None
          )
      with Not_found -> raise Doesnot_fit
    else
      begin let l = tu x x' in
        l @ try
          search cl (fun cl ->
              try
                if mem_var target (snd cl) then
                  let l = unify_pat (fst cl) (fst a) in
                  Some (l @ tu (snd cl) (subst_impl l (snd a)))
                else
                  None
              with
              | Doesnot_fit -> None
            )
        with Not_found -> raise Doesnot_fit end

  | Explicit.Match(_, _), Implicit.Match(_, _) ->
    Format.eprintf "Multi-clause match in patch not supported@."; assert false
  | Explicit.App(a, b), App(a', b') ->
    tu a a' @ tu b b'
  | _, Var((UserId u) as id) when u.[0] = '_' ->
    [id, to_impl_expr e_]
  | Explicit.Var(x, _), Var(i) ->
    if (match i with | Ast.UserId a -> a = x.id_name | Ast.FromId i' -> x = i')
    then []
    else raise Doesnot_fit
  | Explicit.Con(i, _, c), Con(u, l) ->
    if (match u with | Ast.UserId a -> a = i.id_name | Ast.FromId i' -> i = i') then
      List.combine c l
      |> List.map (fun (a, b) -> tu a b)
      |> List.concat
    else raise Doesnot_fit
  | Explicit.Con(x, _, c), Tuple l when List.length l = List.length c ->
    List.combine c l
    |> List.map (fun (a, b) -> tu a b)
    |> List.concat
  | Explicit.MultiFix([fun_name, (v, _), e], _), Let(_, true, [fun_name',  (Abs(v', e'), _)], _) ->
    let e'_renamed = (subst_impl [v', (Var(FromId v), dummy_loc); fun_name', (Var(FromId fun_name), dummy_loc)] e') in
    ( fun_name', (Var(FromId fun_name), dummy_loc)) :: tu e e'_renamed
  | Explicit.Abs(v, _, e), Abs(v', e') ->
    [v', (Var(FromId v), dummy_loc)] @ tu e (subst_impl [v', (Var(FromId v), dummy_loc)] e')
  | Explicit.Let(_, _, _, [], e), _ ->
    tu e e'_
  | _, Let(_, _, [], e') ->
    tu e_ e'
  | Explicit.Let(is_meta, is_rec, ids, bindings, e), Let(is_meta', is_rec', bindings', e') ->
    let n_bindings = List.length bindings in
    let n_bindings' = List.length bindings' in
    if n_bindings <= n_bindings' then
      begin
        let bindings', bindings_next' = chop n_bindings bindings' in
        let l = List.map2 (fun (v, _, e) (v', e') ->
            (v', (Var(FromId v), dummy_loc)) ::tu e e') bindings bindings' |> List.concat in
        l @ tu e (Let(is_meta', is_rec', bindings_next', e'), dummy_loc)
      end
    else
      begin
        let bindings, bindings_next = chop n_bindings' bindings in
        let l = List.map2 (fun (v, _, e) (v', e') ->
            (v', (Var(FromId v), dummy_loc)) ::tu e e') bindings bindings' |> List.concat in
        l @ tu (Let(is_meta, is_rec, ids, bindings_next, e), ty_e, dummy_loc) e'
      end
  | _ -> raise Doesnot_fit
and unify_pat (p_:Explicit.pat) (p'_: Implicit.pat) =
  let p, _, _ = p_ in
  let p', _ = p'_ in
  (match p, p' with
   | Ast.Explicit.PCon (i, _, l), PCon(u, l') ->
     if (match u with | Ast.UserId a -> a = i.id_name | Ast.FromId i' -> i = i') then
       List.combine l l'
       |> List.map (fun (a, b) -> unify_pat a b)
       |> List.concat
     else
       raise Doesnot_fit
   | _, PVar x -> [x, pat_to_impl p_]
   | _, PWild -> []
   | Ast.Explicit.PCon (_, _, l), PTuple p -> (* FIXME: we should check that the constructor is indeed a tuple constructor *)
     if List.length l <> List.length p then raise Doesnot_fit;
     List.combine l p
     |> List.map (fun (a, b) -> unify_pat a b)
     |> List.concat
   | Ast.Explicit.PAlt (a, b), PAlt(a', b') -> unify_pat a a' @ unify_pat b b'
   | _, PAlt(a', b') ->
     begin
       try unify_pat p_ a'
       with Doesnot_fit -> unify_pat p_ b'
     end
   | _ -> (*Format.eprintf "Does not fit pat %a@." Pp.pat p_; Pp.impl_pat_stderr p'_; Format.eprintf "@.";*) raise Doesnot_fit)

let match_expr (env:env) (target_id: id) (stack: (env * expr) list) (e: Ast.expr) (i: patch_addr): (user_id * impl_expr) list option =
  (* first, let's build the in-depth exploded expr view *)
  match Implicit.visit_expr (fun f l e ->
      match e with
      | Implicit.Patch(p), _ when p = i -> f, Some (e::f), e
      | Implicit.Let(_, true, [a, e'], _), _ -> e :: f, l, e'
      | _ -> e::f, l, e) [] None e with
  | Some stack', _ -> (* now we have to compare the two stacks and output variables *)
    begin try
        let n = List.length stack' in
        if List.length stack < n then
          let _ = Format.eprintf "context too small to use this patch@." in None
        else
          let _, expr_code = List.nth stack (n-1) in
          let expr_patch = List.nth stack' (n-1) in (* supposedly the last one *)
          Some (try_unify env target_id expr_code expr_patch)
      with
      | Doesnot_fit -> None
    end
  | None, _ -> assert false

let rm_wilds_from_pat pat =
  Explicit.visit (fun () p -> (),
                              match p with
                              | PWild, ty, loc -> PVar (id "_"), ty, loc
                              | _ -> p) () pat |> snd

let rm_wilds e =
  Explicit.visit_expr (fun () () e ->
      All (), (), match e with
      | Explicit.Match(a, l), ty, loc ->
        let l = List.map (fun (a, b) -> rm_wilds_from_pat a, b) l in
        Explicit.Match(a, l), ty, loc
      | _ -> e ) () () e |> snd

let extract_patch: Implicit.expr -> Implicit.expr * patch_addr * Implicit.expr = fun e ->
  visit_expr (fun _ a_opt e ->
      let a_opt = match e with
        | Implicit.Patch((PatchFor ((_, Some mye):: _)) as pa), _ -> Some(pa, mye)
        | _ -> a_opt in
      (), a_opt, e) () None e |> fst
  |> fun m -> match m with Some (a, b) -> e, a, b | None -> failwith "syntax error, no patch"


let insert_patch: Implicit.expr -> Implicit.expr -> Implicit.expr = fun e e' ->
  visit_expr (fun _ _ e ->
      let e = match e with
        | Implicit.Patch(PatchFor l), loc ->  Implicit.Patch (PatchFor (List.map (fun (a, _) -> a, Some e') l)), loc
        | _ -> e in
      (), (), e) () () e |> snd

let rec subst_in_term2 (oenv: Explicit.orn_env) (env: env) (pl: patch_list) (reqs: Orn.instance_request list) (e: expr): (id * Orn.E.expr) list =
  Explicit.visit_expr (fun (env, stack) l e ->
      let stack = match e with Mark(_), _, _ -> stack | _ -> (env, e) :: stack in
      let next_ctx = begin match e with
        | Explicit.Abs(arg_name, _, _), _, _ ->
          let scope = Scope.intro_named env.scope Scope.NsVar (UserId (arg_name.id_name)) arg_name in
          Explicit.All ({ env with scope }, stack)
        | Explicit.Let(_, _, _, iddef, _), _, _ ->
          let scope =
            List.fold_left (fun scope (arg_name, _, _) ->
                Scope.intro_named scope Scope.NsVar (UserId (arg_name.id_name)) arg_name) env.scope iddef in
          Explicit.All ({ env with scope }, stack)
        | Explicit.MultiFix(l, _), _, _ ->
          let l = List.map (fun (i, (arg_name, _), _) ->
              let scope = Scope.intro_named env.scope Scope.NsVar (UserId (arg_name.id_name)) arg_name in
              { env with scope }, stack
            ) l in
          Explicit.Each l
        | _ -> Explicit.All (env, stack)
      end in
      next_ctx, begin match e with
        | Explicit.Var(i, _), _, _ ->
          begin match List.find (function | Orn.InstantiatePatch(_, _, _, xs, _) -> List.mem i xs | _ -> false) reqs with
            | Orn.InstantiatePatch(vars, _, n, xs, _) ->
              begin match search pl (function
                  | PatchHole(e', a,e), _ when match_addr a n ->
                    begin
                      let e' = Ml_inference.fix_contuple env e' in
                      let subst = match_expr env i stack e' a in
                      match subst with
                      | Some subst ->
                        Some(e', subst, a, e)
                      | None -> None
                    end
                  | _ -> None) with
              | e', subst, a, e ->
                let e = Ml_inference.fix_contuple env e in
                let e = subst_impl (subst) e in
                let e = Orn_codegen.name_patch env.scope [] e in
                let e = Desugaring.untyped_all_patch env e in
                let e = Eml_compute.ExplicitToGml.ml_to_gml e in
                (i, e) :: l
              | exception Not_found ->
                l
              end
            | _ -> assert false
            | exception Not_found -> l
          end
        | _ -> l end, e) (env, []) []  e
  |> fst

let patch_expr (env: env) (pl: patch_list) (e:expr): expr = assert false
