type string = String
type int = Int
type 'a list = Nil | Cons of 'a * 'a list
type 'a option = None | Some of 'a

type 'a box = Box of 'a
type expr = App of expr * expr | Const of int | Abs of string * expr | Var of string box

type bool = True | False

let eq a b = True

val prev: int -> int
val zero: int

let rec assoc n l = match l with
  | Nil -> None
  | Cons((a, b), q) ->
    match n with
    | Box n' ->
      match eq a n' with
      | True -> Some b
      | False -> assoc (Box n') q

let rec eval env e =
  match e with
  | Const i -> Some (Const i)
  | App(a, b) ->
    begin
    match eval env b with
    | None -> None
    | Some e -> begin match eval env a with
      | Some (Abs(n, v)) -> eval (Cons((n, e), env)) v
      | _ -> None
      end
    end
  | Abs(n, v) -> Some (Abs(n, v))
  | Var(n) -> assoc n env

type expr_de_bruijn = App' of expr_de_bruijn * expr_de_bruijn | Const' of int | Abs' of expr_de_bruijn | Var' of int box

type ornament ('a, 'b) erase_box: 'a box => 'b box with
  | Box _ => Box _

type ornament bruijnification: expr => expr_de_bruijn with
  | Var a => Var' a when a: (string, int) erase_box
  | App (a, b) => App'(a, b) when a b: bruijnification
  | Const i => Const' i
  | Abs(_, e) => Abs' e when e: bruijnification

type ornament env: (string * expr) list => expr_de_bruijn list with
  | Nil => Nil
  | Cons((_, a), q) => Cons(a, q) when q: env, a: bruijnification


let assoc' = lifting assoc: (_, _) erase_box -> env -> _ option with
  | #2 <- prev (match n with Box n' -> n')
  | #16 <- zero
  | #10 <- match n with Box a -> a

let eval_bruijinified = lifting eval: env -> bruijnification -> bruijnification option with
  | lifting #5 <- assoc'
  | #32 <- match e with Var' (Box n) -> n

(* not possible to get assoc back *)
let assoc2 = lifting assoc': (_, _) rev_erase_box -> rev_env -> _ option
