type int = Integer

val add: int -> int -> int
val mul: int -> int -> int

type expr =
  | Const of int
  | Add of expr * expr
  | Mul of expr * expr

let rec eval e = match e with
  | Const i -> i
  | Add ( u , v ) -> add (eval u)  (eval v)
  | Mul ( u , v ) -> mul (eval u)  (eval v)

let rec copy e = match e with
  | Const (i) -> Const i
  | Add ( u , v ) ->  Add ( copy u , copy v )
  | Mul ( u , v ) -> Mul ( copy u , copy v )


type binop' =
  | Add'
  | Mul'
  | Sub'

type expr' =
  | Const' of int
  | Binop' of binop' * expr' * expr'

type ornament oexpr : expr => expr' with
  | Const ( i ) => Const' ( i )
  | Add ( u , v ) => Binop' ( Add' , u , v ) when u v : oexpr
  | Mul ( u , v ) => Binop' ( Mul' , u , v ) when u v : oexpr
  | ~ => Binop'(Sub', _, _)

let eval = lifting eval : oexpr -> int
