type nat = | Z
           | S of nat

type 'a list =
  | Nil
  | Cons of 'a * 'a list

let rec append m n = match m with
    | Nil -> n
    | Cons(a, m) -> Cons(a, append m n)

type ornament 'a natlist : nat => 'a list with
  | Z => Nil
  | S xs => Cons(_,xs) when xs : 'a natlist

let add = lifting append: _ rev_natlist -> _ rev_natlist -> _ rev_natlist



type int = Integer
type string = String

type 'a option =
  | Some of 'a
  | None

type expr =
  | Abs of (expr -> expr)
  | App of expr * expr
  | Const of int
type location = Location of string * int * int

type expr_loc =
  | Abs' of (expr_loc -> expr_loc) * location
  | App' of expr_loc * expr_loc * location
  | Const' of int * location

type ('a, 'err) result =
  | Ok of 'a
  | Error of 'err

type ornament add_loc : expr => expr_loc with
  | Abs f => Abs' (f, _)  when f : add_loc -> add_loc
  | App (u, v) => App' (u, v, _) when u v : add_loc
  | Const i => Const' (i, _)

type ornament ('a, 'err) optres : 'a option => ('a, 'err) result with
  | Some a => Ok a
  | None => Error _

let rec eval_loc e = match e with
  | App'(u, v, loc) ->
    begin match eval_loc u with
      | Ok (Abs'(f, _)) -> eval_loc (f v)
      | (Ok (App'(_, _, _)) | Ok (Const'(_, _))) ->
        Error loc
      | Error x -> Error x
    end
  | _ -> Ok e

let eval = lifting eval_loc: rev_add_loc -> (rev_add_loc, location) rev_optres
