type int = Int
type string = String
type 'a option =
  | Some of 'a
  | None

type expr =
  | Abs of (expr -> expr)
  | App of expr * expr
  | Const of int

type location = Location of string * int * int

type expr_loc =
  | Abs' of (expr_loc -> expr_loc) * location
  | App' of expr_loc * expr_loc * location
  | Const' of int * location

type ornament add_loc : expr_loc => expr with
  | Abs' (f, _) => Abs f when f : add_loc -> add_loc
  | App'(u, v, _) => App (u, v) when u v : add_loc
  | Const'(i, _) => Const i

type ('a,'err) result =
  | Ok of 'a
  | Error of 'err

type ornament ('aa, 'err) optres : ('aa, 'err) result => 'aa option with
  | Ok a => Some a
  | Error _ => None

let rec eval_loc e = match e with
  | Abs'(_, _) -> Ok e
  | App'(u, v, loc) ->
    begin match eval_loc u with
      | Ok (Abs'(f, _)) -> eval_loc (f v)
      | Ok (App'(_, _, _)) -> Error loc
      | Ok (Const'(_, _)) -> Error loc
      | Error x -> Error x
    end
  | Const'(_, _) -> Ok e

let c = lifting eval_loc

let e = lifting eval_loc: add_loc -> (add_loc, location) result

let d = lifting eval_loc: add_loc -> (add_loc, location) optres


type expr' =
  | Abs' of (expr' * location -> expr' * location)
  | App' of (expr' * location) * (expr' * location)
  | Const' of int

type ornament add_loc2 : expr' * location => expr with
  | (Abs' f, _) => Abs f when f : add_loc2 -> add_loc2
  | (App' (u, v), _) => App (u, v)  when u v : add_loc2
  | (Const' (i), _) => Const (i)

let rec eval_loc_tuple e = match e with
    | (Abs' _, _) -> Ok e
    | (App'(u, v), loc) ->
      begin match eval_loc u with
        | Ok (Abs' f, _) -> eval_loc (f v)
        | Ok (App'(_, _), _) -> Error loc
        | Ok (Const' _, _) -> Error loc
        | Error x -> Error x
      end
    | (Const' _, _) -> Ok e


let e = lifting eval_loc: add_loc -> _ optres
