type 'a box = Box of 'a

type bool = True | False

type 'a option = None | Some of 'a

val i0 : int
val i1 : int
val i2 : int
val gt : int -> int -> bool
val gte : int -> int -> bool
val plus : int -> int -> int

type cmp = Eq | Lt | Gt

type elt = Elt
val compare: elt -> elt -> cmp


type 'a map =
  | MEmpty
  | MNode of 'a map * elt * 'a * 'a map * int

  type 'a map = MEmpty | MNode of 'a map * elt * 'a * 'a map * int

  type t = Empty | Node of t * elt * t * int

  type ornament 'a setmap : 'a map => t with
       | MEmpty => Empty
       | MNode(l,v,_,r,n) => Node(l,v,r,n) when l r : 'a setmap

  type ornament ('a,'new) box2 : ('a * 'new) box => 'a box with
       | Box (a,_) => Box a

  type ornament ('a, 'b) option2 : ('a * 'b) option => 'a option  with
       | None => None
       | Some (u, _) => Some u


let minvalid_arg = MEmpty

let invalid_arg = lifting minvalid_arg : _ setmap


let mheight t = match t with
    | MEmpty -> i0
    | MNode(_, _, _, _, h) -> h

let height = lifting mheight : _ setmap -> _

let mcreate l vb r = match vb with
    | Box (a, x) ->
      let hl = match l with
          | MEmpty -> i0
          | MNode(_, _, _, _, n) -> n
      in
      let hr = match r with
          | MEmpty -> i0
          | MNode(_, _, _, _, n) -> n
      in
      let x1 = (* _2 *) gte hl hr in
      MNode(l,
       a,
       x,
       r,
       match x1 with
         | True -> plus hl i1
         | False -> plus hr i1
      )

      let create = lifting mcreate : _ setmap -> _ box2 -> _ setmap -> _ setmap
let mbal l vb r = match vb with
    | Box (a, x) ->
      let hl = match l with
          | MEmpty -> i0
          | MNode(_, _, _, _, n) -> n
      in
      let hr = match r with
          | MEmpty -> i0
          | MNode(_, _, _, _, n) -> n
      in
      match gt hl (plus hr i2) with
        | True ->
          begin match l with
            | MEmpty -> minvalid_arg
            | MNode(l, v, x1, r1, _) ->
              begin match gte (mheight l) (mheight r1) with
                | True ->
                  mcreate l (Box (v, x1))
                  (mcreate r1 (Box (a, x)) r)
                | False ->
                  begin match r1 with
                    | MEmpty -> minvalid_arg
                    | MNode(l1, v1, x2, r1, _) ->
                      mcreate (mcreate l (Box (v, x1)) l1)
                      (Box (v1, x2))
                      (mcreate r1 (Box (a, x)) r)
                  end
              end
          end
        | False ->
          begin match gt hr (plus hl i2) with
            | True ->
              begin match r with
                | MEmpty -> minvalid_arg
                | MNode(l1, v, x1, r, _) ->
                  begin match gte (mheight r) (mheight l1)
                    with
                    | True ->
                      mcreate (mcreate l (Box (a, x)) l1)
                      (Box (v, x1))
                      r
                    | False ->
                      begin match l1 with
                        | MEmpty -> minvalid_arg
                        | MNode(l1, v1, x2, r1, _) ->
                          mcreate
                          (mcreate l (Box (a, x)) l1)
                          (Box (v1, x2))
                          (mcreate r1 (Box (v, x1)) r)
                      end
                  end
              end
            | False ->
              let x1 = (* _4 *) gte hl hr in
              MNode(l,
               a,
               x,
               r,
               match x1 with
                 | True -> plus hl i1
                 | False -> plus hr i1
              )
          end

let rec madd xb t = match xb with
    | Box (a, x) ->
      begin match t with
        | MEmpty -> MNode(MEmpty, a, x, MEmpty, i1)
        | MNode(l, v, x1, r, _) ->
          begin match compare a v with
            | Eq -> t
            | Lt ->
              mbal (madd (Box (a, x)) l) (Box (v, x1)) r
            | Gt ->
              mbal l (Box (v, x1)) (madd (Box (a, x)) r)
          end
      end

let msingleton xb = match xb with
    | Box (a, x) -> MNode(MEmpty, a, x, MEmpty, i1)

let rec mmin_binding t = match t with
    | MEmpty -> None
    | MNode(MEmpty, u, x, _, _) -> Some (u, x)
    | MNode(MNode(l, v, x, r, n), _, _, _, _) ->
      mmin_binding (MNode(l, v, x, r, n))

let rec mmax_binding t = match t with
    | MEmpty -> None
    | MNode(_, u, x, MEmpty, _) -> Some (u, x)
    | MNode(_, _, _, MNode(l, v, x, r, n), _) ->
      mmax_binding (MNode(l, v, x, r, n))

let bal = lifting mbal : _ setmap -> _ box2 -> _ setmap -> _ setmap

let add = lifting madd : _ box2 -> _ setmap -> _ setmap

let singleton = lifting msingleton : _ box2 -> _ setmap


let min_binding = lifting mmin_binding : _ setmap -> _ option2

let max_binding = lifting mmax_binding : _ setmap -> _ option2
