type int = Z | S of int | Sn of int
type 'a list = Nil | Cons of 'a * 'a list

(** sorting *)

let rec length_aux len = function
    Nil -> len
  | Cons(_, l) -> length_aux (S len) l

let length l = length_aux Z l

let rec rev_append l1 l2 =
  match l1 with
    Nil -> l2
  | Cons(a, l) -> rev_append l (Cons(a, l2))

let rec merge cmp l1 l2 =
  match (l1, l2) with
  | (Nil, l2) -> l2
  | (l1, Nil) -> l1
  | (Cons(h1, t1), Cons(h2, t2)) ->
      match cmp h1 h2 with
      | S _ -> Cons(h2, merge cmp l1 t2)
      | _ -> Cons(h1, merge cmp t1 l2)


let rec chop k l =
  match k with
  | Z -> l
  | S c -> begin
    match l with
    | Cons(_, t) -> chop c t
    | _ -> (* should not happen *) l
  end
  | _ -> (* should not happen *) l

val aasr: int -> int -> int
val minus: int -> int -> int

let stable_sort cmp l =
  let rec rev_merge l1 l2 accu =
    match (l1, l2) with
    | (Nil, l2) -> rev_append l2 accu
    | (l1, Nil) -> rev_append l1 accu
    | ((Cons( h1, t1)) , (Cons( h2, t2 ))) ->
        match cmp h1 h2 with
        | Sn _ -> rev_merge t1 l2 ( (Cons(h1, accu)) )
        | _ -> rev_merge l1 t2 ( (Cons(h2, accu)) )
  in
  let rec rev_merge_rev l1 l2 accu =
    match (l1, l2) with
    | (Nil, l2) -> rev_append l2 accu
    | (l1, Nil) -> rev_append l1 accu
    | ((Cons( h1, t1)) , (Cons( h2, t2 ))) ->
        match cmp h1 h2 with
        | S _ -> rev_merge_rev t1 l2 ( (Cons(h1, accu)) )
        | _ ->  rev_merge_rev l1 t2 ( (Cons(h2, accu)) )
  in
  let rec sort n l =
       let n1 = aasr n (S Z) in
       let n2 = minus n n1 in
       let l2 = chop n1 l in
       let s1 = rev_sort n1 l in
       let s2 = rev_sort n2 l2 in
       rev_merge_rev s1 s2 Nil
  and rev_sort n l =
       let n1 = aasr n (S Z) in
       let n2 = minus n n1 in
       let l2 = chop n1 l in
       let s1 = sort n1 l in
       let s2 = sort n2 l2 in
       rev_merge s1 s2 Nil
  in
  let len = length l in
  sort len l

  type ornament 'a natlist : nat => 'a list with
    | Z => Nil
    | S xs => Cons(_,xs) when xs : 'a natlist

let stable_sort' = lifting stable_sort: _ -> 'a rev_natlist -> _
