
type 'a option =
  | Some of 'a
  | None

type expr =
  | Abs of (expr -> expr)
  | App of expr * expr
  | Const of int

let rec eval e = match e with
  | App (u, v) ->
      begin match eval u with Some (Abs f) -> eval (f v) | _ -> None end
  | _ -> Some(e)

type location = Location of string * int * int
type expr_loc =
  | Abs' of (expr_loc -> expr_loc) * location
  | App' of expr_loc * expr_loc * location
  | Const' of int * location

type ('a, 'err) result =
  | Ok of 'a
  | Error of 'err

type ornament add_loc : expr => expr_loc with
  | Abs f => Abs' (f, _)  when f : add_loc -> add_loc
  | App (u, v) => App' (u, v, _) when u v : add_loc
  | Const i => Const' (i, _)

type ornament ('a, 'err) optres : 'a option => ('a, 'err) result with
  | Some a => Ok a
  | None => Error _


let eval_incomplete = lifting eval : add_loc -> (add_loc, location) optres

let eval_loc =  lifting eval : add_loc -> (add_loc, location) optres with
  | #6, #8 <- (match e with App'(_,_,loc) -> loc)
  | (match _a with Error x -> Error (#a)) => #a <- x

type expr' =
  | Abs' of (expr' * location -> expr' * location)
  | App' of (expr' * location) * (expr' * location)
  | Const' of int

type ornament add_loc : expr => expr' * location with
  | Abs f => (Abs' f, _) when f : add_loc -> add_loc
  | App (u, v) => (App' (u, v), _) when u v : add_loc
  | Const (i) => (Const' (i), _)

let eval_incomplete = lifting eval : add_loc -> _ optres

let eval' = lifting eval: add_loc -> (_, _) optres with
| (match _a with Error x -> Error (#a)) => #a <- x
| (match _a with Ok _ -> Error (#a)) => #a <- (match e with (_, loc) -> loc)
