type int = Integer

type string = String

type 'a option =
  | Some of 'a
  | None

type expr =
  | Abs of (expr -> expr)
  | App of expr * expr
  | Const of int
  | Pouf of string

type location = Location of string * int * int

type expr' =
  | Abs' of ((expr' * location) -> (expr' * location))
  | App' of (expr' * location) * (expr' * location)
  | Const' of int
  | Pouf' of string

type relation add_loc : expr => (expr' * location) with
 | Abs f => (Abs' f, _) when f:(add_loc -> add_loc)
 | App (u, v) => (App' (u, v), _) when u:add_loc,v:add_loc
 | Const i => (Const' i, _)
 | Pouf _ => (Pouf' _, _)

type ('a,'err) result =
  | Ok of 'a
  | Error of 'err

type relation ('a,'err) optres : 'a option => ('a,'err) result with
 | Some a => Ok a
 | None => Error _

let rec eval_loc e = match e with
    | (App'(u, v), x) ->
      begin match eval_loc u with
        | Ok (Abs' f, _) -> eval_loc (f v)
        | (Ok (App'(_, _), _) | Ok (Const' _, _)) -> Error x
        | Ok (Pouf' _, _) -> Error x
        | Error x -> Error x
      end
    | (Pouf' a, x) ->
    	Ok (Pouf' a, x)
    | _ -> Ok e

#(mode genpatches)
let eval = lifting eval_loc : (rev_add_loc -> (rev_add_loc,location) rev_optres)
with #6 <- match e with Pouf a -> a
