type nat = Z | S of nat

let rec add x y =
  match y with
  | Z -> x
  | S z -> S(add x z)

let rec id_nat y = 
  match y with
  | Z -> Z
  | S z -> S (id_nat z)

type bool = True | False
type num = Z' | S' of bool * num

type ornament nat_num : nat => num with
  | Z => Z'
  | S z => S' (_, z) when z : nat_num

let id_num = lifting id_nat : nat_num -> nat_num
   with #2 <- (match y with S' (x1, _) -> x1)

let num_to_nat = lifting id_nat : nat_num -> nat
    
let add_num = lifting add : nat_num -> nat_num -> nat_num
   with #2 <- (match y with S' (x1, _) -> x1)

type num2 = Z0 | S1 of num2 | S2 of num2
    
type ornament num_num2 : num => num2 with
  | Z' => Z0
  | S' (True, z) => S1 z when z : num_num2
  | S' (False, z) => S2 z when z : num_num2

let num2_to_num = lifting id_num : num_num2 -> num

let num2_to_nat x = num_to_nat (num2_to_num x)

        
let add_num2 = lifting add_num : num_num2 -> num_num2 -> num_num2 

(* the following is accepted, but not correctly interpreted *)
type ornament nat_num2 : nat => num2 with
  | Z => Z0
  | S z => S1 z when z : nat_num2
  | S z => S2 z when z : nat_num2

(* this should be correct *)
type ornament nat_num2 : nat => num2 with
  | Z => Z0
  | S z => (S1 z | S2 z) when z : nat_num2

(* Since the following is not flag ambiguous *)
let add_num2' = lifting add : nat_num2 -> nat_num2 -> nat_num2 

(* The solution should be to treat contruction of a S_skel as ambiguous : 
   the user should choose  between the constructors S1 and S2. 
   This would still be less ambiguous that for add_num since only the 
   constructor has to be choosen, ie. the incomplete code could show an 
   or pattern (S1 z | S2 z) in the expression side
 *)

(* The converse ornament that identifies two constructors is defined 
   as follows *) 


    
type ornament num2_nat : num2 => nat with
  | Z0 => Z
  | S1 z => S z when z : num2_nat
  | S2 z => S z when z : num2_nat

(* It works as expected on add_mum2... *)

let add' = lifting add_num2 : num2_nat -> num2_nat -> num2_nat 

(* but this is by change, as it will just ignores the Z2 branch 
   in the lifting, as show below.   *)

let rec add_num2'' x y = match y with
    | Z0 -> x
    | S1 z -> S1 (add_num2 x z)
    | S2 z -> x

let add'' = lifting add_num2'' : num2_nat -> num2_nat -> num2_nat 
          
(* Shouldn't there be a warning that the ornament is not injective? *)

(* If we do it in two steps... *)

type ornament num2_num : num2 => num with
  | Z0 => Z'
  | S1 z => S' (True, z) when z : num2_num
  | S2 z => S' (False, z) when z : num2_num
    
let add_num' = lifting add_num2 : num2_num -> num2_num -> num2_num
let add_num'' = lifting add_num2'' : num2_num -> num2_num -> num2_num

(* --rejeted as non-injective 
type ornament num_nat : num => nat with
  | Z' => Z
  | S' (x, z) => S z when z : num_nat
*)
    
(* let add' = lifting add_num' : num_nat -> num_nat -> num_nat  *)

type a = A | B
type c = C

(* non injectif, accepté par erreur *)
type ornament ab : a => c with
  | A => C
  | B => C

(* We have to chose which side to throw, for example *)
type ornament ab : a => c with
  | A => ~
  | B => C

let a_to_nat x = match x with
  | A -> Z
  | B -> S Z

let c_to_nat = lifting a_to_nat : ab -> nat 
      
      
type ornament num_nat : num => nat with
  | Z' => Z
  | S' (True, z) => S z when z : num_nat
  | S' (False, z) => ~ when z : num_nat

let add_xxx = lifting add_num : num_nat -> num_nat -> 'nat
