type int = Integer
type 'a option = Some of 'a | None
val add: int -> int -> int
val mul: int -> int -> int

type expr =
| Const of int
| Add of expr * expr
| Mul of expr * expr
| Abs of (expr -> expr option)
| App of expr * expr
let rec eval e = match e with
| Const i -> Some(Const i)
| Add ( u , v ) ->
   begin match (eval u, eval v) with
   | (Some (Const i1), Some (Const i2)) -> Some(Const (add i1 i2))
   | _ -> None
   end
| Mul ( u , v ) ->
   begin match (eval u, eval v) with
   | (Some (Const i1), Some (Const i2)) -> Some(Const (mul i1 i2))
   | _ -> None
   end
| Abs f -> Some(Abs f)
| App(u, v) ->
   begin match eval u with
   | Some(Abs f) -> begin match eval v with None -> None | Some x -> f x end
   | _ -> None
   end
type binop' =
  | Add'
  | Mul'

type expr' =
  | Const' of int
  | Binop' of binop' * expr' * expr'
  | Abs' of (expr' -> expr' option)
  | App' of expr' * expr'
type ornament oexpr : expr => expr' with
  | Const ( i ) => Const' ( i )
  | Add ( u , v ) => Binop' ( Add' , u , v ) when u v : oexpr
  | Mul ( u , v ) => Binop' ( Mul' , u , v ) when u v : oexpr
  | Abs f => Abs' f when f : oexpr -> oexpr option
  | App ( u , v ) => App' ( u , v ) when u v : oexpr

let u x y = let u = (eval x, eval y) in match (eval x, eval y) with | (a, b) -> (a, b)

let u' = lifting u with ornament * <- @id
