#(mode elimdeadcode)

type elt = Abstract_element

type t =
  | Black of t * elt * t
  | Red of t * elt * t
  | Empty

type enum =
  | End
  | More of elt * t * enum

type bool = True | False
type res_cmp = Lt | Gt | Eq

val compare : elt -> elt -> res_cmp

let rec enum s e =
  match s with
    | Empty -> e
    | (Black(l, x, r) | Red(l, x, r)) -> enum l (More(x, r, e))

let rec mem x = function
  | Empty ->
      False
  | (Red(l, y, r)
  | Black(l, y, r)) ->
      match compare x y with
      | Lt -> mem x l
      | Gt -> mem x r
      | Eq -> True

let balance_left l x r =
  match (l, x, r) with
    | ((Red(Red(a, x, b), y, c), z, d)
    | (Red(a, x, Red(b, y, c)), z, d)) ->
        Red(Black(a, x, b), y, Black(c, z, d))
    | (l, x, r) ->
        Black(l, x, r)

let balance_right l x r =
  match (l, x, r) with
    | ((a, x, Red(Red(b, y, c), z, d))
    | (a, x, Red(b, y, Red(c, z, d)))) ->
        Red(Black(a, x, b), y, Black(c, z, d))
    | (l, x, r) ->
        Black(l, x, r)

let add x s =
  let rec add_aux t = match t with
    | Empty ->
        Red(Empty, x, Empty)
    | Red(l, y, r) ->
        begin match compare x y with
        | Lt -> Red(add_aux l, y, r)
        | Gt -> Red(l, y, add_aux r)
        | Eq -> t end
    | Black(l, y, r)  ->
        begin match compare x y with
        | Lt -> balance_left (add_aux l) y r
        | Gt -> balance_right l y (add_aux r)
        | Eq -> t
        end
  in match ((add_aux s)) with
  | Red(a, b, c) -> Black(a, b, c)
  | a -> a

let singleton x =
Black(Empty, x, Empty)

type tree =  SEmpty | SNode of tree * elt * tree
(*type rb =  REmpty | RNode of bool * rb * elt * rb

type ornament booleanize: t => rb with
| Empty => REmpty
| Red(a, e, b) => RNode(True, a, e, b) when a b: booleanize
| Black(a, e, b) => RNode(False, a, e, b) when a b: booleanize


let brright = lifting balance_right: booleanize -> elt -> booleanize -> booleanize
with ornament * <- @id

let brleft = lifting balance_left: booleanize -> elt -> booleanize -> booleanize
with ornament * <- @id

let brblacki = lifting blackify: booleanize -> booleanize * bool
with ornament * <- @id

let bradd = lifting add: elt -> booleanize -> booleanize

type ornament simplification: rb => tree with
| REmpty => SEmpty
| RNode(_, a, e, b) => SNode(a, e, b) when a b: simplification*)

type relation simplification: t => tree with
| Empty => SEmpty
| (Red(a, e, b) | Black(a, e, b)) => SNode(a, e, b) when a b: simplification

(*let br2 = lifting brright: simplification -> elt -> simplification -> simplification
with ornament * <- @id
| #a <- True

let brmem = lifting mem: elt -> booleanize -> bool with #a <- True
let brmem = lifting brmem: elt -> simplification -> bool with #a <- True | ornament * <- @id
let bradd = lifting bradd: elt -> simplification -> simplification with #a <- True*)


let simpl_mem = lifting mem: elt -> simplification -> bool

let simpl_add = lifting add: elt -> simplification -> simplification

let simpl_add = lifting add: elt -> simplification -> simplification with
| (match #a with Left _ -> _a)
  => #a <- Left ()


let simpl_singleton = lifting singleton: elt -> simplification
